@@ -2,19 +2,18 @@
  * Copyright (C) 2001,2002 Red Hat, Inc.
  * Copyright © 2009, 2010, 2019 Christian Persch
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 3 of the License, or (at your option) any later version.
+ * This library is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as published
+ * by the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this library.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 /**
@@ -152,7 +151,7 @@ catch (...)
  * @error: (allow-none): return location to store a #GError, or %NULL
  *
  * Attempts to resize the pseudo terminal's window size.  If successful, the
- * OS kernel will send #SIGWINCH to the child process group.
+ * OS kernel will send <literal>SIGWINCH</literal> to the child process group.
  *
  * If setting the window size failed, @error will be set to a #GIOError.
  *
@@ -504,8 +503,9 @@ vte_pty_error_quark(void) noexcept
  * Note that you should set the PTY's size using vte_pty_set_size() before
  * spawning the child process, so that the child process has the correct
  * size from the start instead of starting with a default size and then
- * shortly afterwards receiving a SIGWINCH signal. You should prefer
- * using vte_terminal_pty_new_sync() which does this automatically.
+ * shortly afterwards receiving a <literal>SIGWINCH</literal> signal. You
+ * should prefer using vte_terminal_pty_new_sync() which does this
+ * automatically.
  *
  * Returns: (transfer full): a new #VtePty, or %NULL on error with @error filled in
  */
@@ -684,7 +684,7 @@ try
                                             cancellable};
 
         auto err = vte::glib::Error{};
-        auto rv = op.run_sync(child_pid, err);
+        auto rv = vte::base::SpawnOperation::run_sync(op, child_pid, err);
         if (!rv)
                 err.propagate(error);
 
@@ -731,12 +731,12 @@ _vte_pty_check_envv(char const* const* strv) noexcept
  * @n_map_fds: the number of elements in @map_fds, or 0 if @map_fds is %NULL
  * @spawn_flags: flags from #GSpawnFlags
  * @child_setup: (allow-none) (scope async): an extra child setup function to run in the child just before exec(), or %NULL
- * @child_setup_data: (closure child_setup): user data for @child_setup, or %NULL
- * @child_setup_data_destroy: (destroy child_setup_data): a #GDestroyNotify for @child_setup_data, or %NULL
+ * @child_setup_data: (nullable) (closure child_setup): user data for @child_setup, or %NULL
+ * @child_setup_data_destroy: (nullable) (destroy child_setup_data): a #GDestroyNotify for @child_setup_data, or %NULL
  * @timeout: a timeout value in ms, -1 for the default timeout, or G_MAXINT to wait indefinitely
  * @cancellable: (allow-none): a #GCancellable, or %NULL
  * @callback: (nullable) (scope async): a #GAsyncReadyCallback, or %NULL
- * @user_data: (closure callback): user data for @callback
+ * @user_data: (nullable) (closure callback): user data for @callback
  *
  * Starts the specified command under the pseudo-terminal @pty.
  * The @argv and @envv lists should be %NULL-terminated.
@@ -755,12 +755,10 @@ _vte_pty_check_envv(char const* const* strv) noexcept
  * This function will take ownership of the file descriptors in @fds;
  * you must not use or close them after this call. All file descriptors in @fds
  * must have the FD_CLOEXEC flag set on them; it will be unset in the child process
- * before calling exec.
- *
- * Note that all  open file descriptors apart from those mapped as above
- * will be closed in the child. (If you want to keep some other file descriptor
- * open for use in the child process, you need to use a child setup function
- * that unsets the FD_CLOEXEC flag on that file descriptor manually.)
+ * before calling man:execve(2). Note also that no file descriptor may be mapped
+ * to stdin, stdout, or stderr (file descriptors 0, 1, or 2), since these will be
+ * assigned to the PTY. All open file descriptors apart from those mapped as above
+ * will be closed when execve() is called.
  *
  * Beginning with 0.60, and on linux only, and unless %VTE_SPAWN_NO_SYSTEMD_SCOPE is
  * passed in @spawn_flags, the newly created child process will be moved to its own
@@ -797,10 +795,11 @@ try
         g_return_if_fail(argv[0] != nullptr);
         g_return_if_fail(envv == nullptr || _vte_pty_check_envv(envv));
         g_return_if_fail(n_fds == 0 || fds != nullptr);
-        for (auto i = int{0}; i < n_fds; ++i)
+        for (auto i = 0; i < n_fds; ++i)
                 g_return_if_fail(vte::libc::fd_get_cloexec(fds[i]));
         g_return_if_fail(n_fd_map_to == 0 || fd_map_to != nullptr);
-        g_return_if_fail(n_fds >= n_fd_map_to);
+        for (auto i = 0; i < n_fd_map_to; ++i) /* Invalid and stdin/out/err not allowed */
+                g_return_if_fail(fd_map_to[i] < -1 || fd_map_to[i] > 2);
         g_return_if_fail((spawn_flags & ~all_spawn_flags()) == 0);
         g_return_if_fail(!child_setup_data || child_setup);
         g_return_if_fail(!child_setup_data_destroy || child_setup_data);
@@ -814,23 +813,24 @@ try
         g_warn_if_fail((spawn_flags & forbidden_spawn_flags()) == 0);
         spawn_flags = GSpawnFlags(spawn_flags & ~forbidden_spawn_flags());
 
-        auto op = new vte::base::SpawnOperation{spawn_context_from_args(pty,
-                                                                        working_directory,
-                                                                        argv,
-                                                                        envv,
-                                                                        fds, n_fds,
-                                                                        fd_map_to, n_fd_map_to,
-                                                                        spawn_flags,
-                                                                        child_setup,
-                                                                        child_setup_data,
-                                                                        child_setup_data_destroy),
-                                                timeout,
-                                                cancellable};
+        auto op = std::make_unique<vte::base::SpawnOperation>
+                (spawn_context_from_args(pty,
+                                         working_directory,
+                                         argv,
+                                         envv,
+                                         fds, n_fds,
+                                         fd_map_to, n_fd_map_to,
+                                         spawn_flags,
+                                         child_setup,
+                                         child_setup_data,
+                                         child_setup_data_destroy),
+                 timeout,
+                 cancellable);
 
-        /* takes ownership of @op */
-        op->run_async((void*)vte_pty_spawn_async, /* tag */
-                      callback,
-                      user_data);
+        vte::base::SpawnOperation::run_async(std::move(op),
+                                             (void*)vte_pty_spawn_async, /* tag */
+                                             callback,
+                                             user_data);
 }
 catch (...)
 {
@@ -850,12 +850,12 @@ catch (...)
  *   variables to be added to the environment before starting the process, or %NULL
  * @spawn_flags: flags from #GSpawnFlags
  * @child_setup: (allow-none) (scope async): an extra child setup function to run in the child just before exec(), or %NULL
- * @child_setup_data: (closure child_setup): user data for @child_setup, or %NULL
- * @child_setup_data_destroy: (destroy child_setup_data): a #GDestroyNotify for @child_setup_data, or %NULL
+ * @child_setup_data: (nullable) (closure child_setup): user data for @child_setup, or %NULL
+ * @child_setup_data_destroy: (nullable) (destroy child_setup_data): a #GDestroyNotify for @child_setup_data, or %NULL
  * @timeout: a timeout value in ms, -1 for the default timeout, or G_MAXINT to wait indefinitely
  * @cancellable: (allow-none): a #GCancellable, or %NULL
  * @callback: (nullable) (scope async): a #GAsyncReadyCallback, or %NULL
- * @user_data: (closure callback): user data for @callback
+ * @user_data: (nullable) (closure callback): user data for @callback
  *
  * Like vte_pty_spawn_with_fds_async(), except that this function does not
  * allow passing file descriptors to the child process. See vte_pty_spawn_with_fds_async()
